/**
 * Versuch Reliable Segment, dass Rechnerabsturz ueberlebt
 * @author <A href="http://www.infnet.verein.de/~rw/">Rudolf Weber</A>
 * Literatur: Gray Reuter, Transaction Processing Kap. 3.7.2
 *            Vorlesung Meyer Wegner, Aufbau von Datenverwaltungsystemen SS 92
 */
#include<time.h>
#include<stdint.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/uio.h>
#include<fcntl.h>
#include<unistd.h>
#include<string.h>
#include<RAMS/elementartypes.h>
#include"pmseg.h"

/**
 * Management of consistency checks
 * persistent structure
 */
struct VerwData_t
{
  /// Unixzeit
  time_t stamp;
  /// Checksumme
  uint32_t checksum;
};

/* transient data */
struct TVerwData_t
{
  bool dirty;
  VerwData_t pverw;
};
const int maxblocks=8;
/**
 * Fr mprotect brauchen wir die Seitenausrichtung.
 * Damit ist zwar das Schreiben nicht ganz atomar, aber ein Fehler
 * resultiert dann in einem Block, der unvollstndig ist und wird
 * somit erkannt.
 */
const size_t BLOCKSIZE=PAGESIZE+sizeof(VerwData_t);
const size_t RELPAGESZ=PAGESIZE;

/**
 * Verschiedene Strategien sind denkbar:
 * z.B. Twinblockstrategie, verschiedene Dateien
 * diese Klasse abstrahiert davon
 */
class ReliableSegStrategy
{
public:

  /**
   * @brief   Oeffnen aller Dateien
   * @returns 1 : if new
   *          0 : if old
   *       abort: if error
   */
  virtual int open(char *filename)=0;
  /// schliessen aller Dateien
  virtual void close()=0;
  /// Redundanzgrad
  virtual int nplex()=0;
  /// Dateidescriptor fuer Replicat i
  virtual int fd(int i)=0;
  /// Offset fr Seitennummer pageno und Replikat i
  virtual off_t offset(int i,int pageno)=0;
};

/**
 * @brief Twinblockstrategie
 *
 * Jeder Block wird nplex 2 mal hintereinander gespeichert
 */
class TwinblockStrategy : public ReliableSegStrategy
{
  int f;
public:
  TwinblockStrategy() {}
  int open(char *filename);
  void close();
  int nplex();
  int fd(int i);
  off_t offset(int i,int pageno);
};

/**
 * @brief Reliable Segment
 *
 *
 */
class ReliableSegment: public PMSeg
{
protected:
  int fd0;
  TVerwData_t vdata[maxblocks];
  ReliableSegStrategy *strat;
public:
  ReliableSegment(ReliableSegStrategy *s);
  static uint32_t calc_checksum(octet_t *page);
  /**
   * @brief bringt Seite pageno ein
   *
   * Die Seite wird mehrfach geschrieben. Sollte ein fehler auftreten,
   * so geht maximal eine Seite kaputt ...
   */
  void write_page(int pageno);
  /**
   * @brief schnelles lesen
   *
   * lesen des erstbesten Replicats. Sollte dies Fehlschlagen, wird
   * read_page_reliable gerufen
   */
  void read_page_optimistic(int pageno);

  /**
   * @brief zuverlaessiges lesen
   *
   * Alle Replikate werden gelesen und abgeglichen
   */
  void read_page_reliable(int pageno);
  /**
   * @brief oeffnen laut Strategie
   * @returns 1 : if new
   *          0 : if old
   *       abort: if error
   */
  int open(char *filename) { return strat->open(filename); }

  /**
   * Handle for pmseg
   */
   virtual void handle(RAMS::PageNo_t pageno,void *addr);

  /**
   * save dirty pages
   */
  void save();
  /**
   * schliessen
   */
  void close() { strat->close(); }
};




